@@ -1,19 +1,40 @@
+# (C) Copyright 2007-2019 Enthought, Inc., Austin, TX
+# All rights reserved.
+#
+# This software is provided without warranty under the terms of the BSD
+# license included in LICENSE.txt and may be redistributed only
+# under the conditions described in the aforementioned license.  The license
+# is also available online at http://www.enthought.com/licenses/BSD.txt
+# Thanks for using Enthought open source!
 """ This code has been inspired from the IPython repository
 
 https://github.com/ipython/ipython/blob/2.x/examples/Embedding/internal_ipkernel.py
 
 """
-from distutils.version import StrictVersion as Version
-import sys
+from __future__ import absolute_import, print_function, unicode_literals
 
-import ipykernel
-from ipykernel.connect import connect_qtconsole
-from ipykernel.kernelapp import IPKernelApp
-from tornado import ioloop
+import atexit
 
+import ipykernel.connect
+import six
+
+from envisage.plugins.ipython_kernel.kernelapp import IPKernelApp
 from traits.api import Any, HasStrictTraits, Instance, List
 
-NEEDS_IOLOOP_PATCH = Version(ipykernel.__version__) >= Version('4.7.0')
+
+# Allow unregistration of atexit handlers registered by IPython machinery.
+
+if six.PY2:
+    def atexit_unregister(func):
+        # Replace the contents, not the list itself, in case anyone else
+        # is keeping references to it. Also use 'not thing == func' instead
+        # of 'thing != func' to match the semantics of the Python 3 code.
+        atexit._exithandlers[:] = list(
+            handler for handler in atexit._exithandlers
+            if not handler[0] == func
+        )
+else:
+    from atexit import unregister as atexit_unregister
 
 
 def gui_kernel(gui_backend):
@@ -56,12 +77,6 @@ class InternalIPKernel(HasStrictTraits):
     #: This is a list of tuples (name, value).
     initial_namespace = List()
 
-    #: sys.stdout value at time kernel was started
-    _original_stdout = Any()
-
-    #: sys.stderr value at time kernel was started
-    _original_stderr = Any()
-
     def init_ipkernel(self, gui_backend):
         """ Initialize the IPython kernel.
 
@@ -71,31 +86,19 @@ class InternalIPKernel(HasStrictTraits):
           The GUI mode used to initialize the GUI mode. For options, see
           the `ipython --gui` help pages.
         """
-        # The IPython kernel modifies sys.stdout and sys.stderr when started,
-        # and doesn't currently provide a way to restore them. So we restore
-        # them ourselves at shutdown.
-        self._original_stdout = sys.stdout
-        self._original_stderr = sys.stderr
-
         # Start IPython kernel with GUI event loop support
         self.ipkernel = gui_kernel(gui_backend)
 
-        # Since ipykernel 4.7, the io_loop attribute of the kernel is not
-        # initialized anymore
-        # Reference: https://github.com/enthought/envisage/issues/107
-        # Workaround: Retrieve the kernel on the IPykernelApp and set the
-        # io_loop without starting it!
-        if NEEDS_IOLOOP_PATCH and not hasattr(self.ipkernel.kernel, 'io_loop'):
-            self.ipkernel.kernel.io_loop = ioloop.IOLoop.instance()
-
         # This application will also act on the shell user namespace
         self.namespace = self.ipkernel.shell.user_ns
         self.namespace.update(dict(self.initial_namespace))
 
     def new_qt_console(self):
         """ Start a new qtconsole connected to our kernel. """
-        console = connect_qtconsole(self.ipkernel.connection_file,
-                                    argv=['--no-confirm-exit'])
+        console = ipykernel.connect.connect_qtconsole(
+            self.ipkernel.connection_file,
+            argv=['--no-confirm-exit'],
+        )
         self.consoles.append(console)
         return console
 
@@ -103,33 +106,23 @@ class InternalIPKernel(HasStrictTraits):
         """ Kill all existing consoles. """
         for c in self.consoles:
             c.kill()
+            c.wait()
             c.stdout.close()
             c.stderr.close()
         self.consoles = []
 
     def shutdown(self):
-        """ Shutdown the kernel.
+        """ Shut the kernel down.
 
         Existing IPython consoles are killed first.
         """
         if self.ipkernel is not None:
             self.cleanup_consoles()
-            self.ipkernel.shell.exit_now = True
-            self.ipkernel.cleanup_connection_file()
-
-            # The stdout and stderr streams created by the kernel use the
-            # IOPubThread, so we need to close them and restore the originals
-            # before we shut down the thread. Without this, we get obscure
-            # errors of the form "TypeError: heap argument must be a list".
-            kernel_stdout = sys.stdout
-            if kernel_stdout is not self._original_stdout:
-                sys.stdout = self._original_stdout
-                kernel_stdout.close()
-
-            kernel_stderr = sys.stderr
-            if kernel_stderr is not self._original_stderr:
-                sys.stderr = self._original_stderr
-                kernel_stderr.close()
-
-            self.ipkernel.iopub_thread.stop()
+            self.ipkernel.close()
+            # ipkernel.close is only registered for ipykernel 5.1.2 and later,
+            # but unregistering something that wasn't registered is safe.
+            atexit_unregister(self.ipkernel.close)
             self.ipkernel = None
+
+            # Remove stored singleton to facilitate garbage collection.
+            IPKernelApp.clear_instance()
